@@ -35,6 +35,7 @@
 
 #include "bftw.h"
 #include "dstring.h"
+#include "stat.h"
 #include "util.h"
 #include <assert.h>
 #include <dirent.h>
@@ -45,8 +46,6 @@
 #include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/stat.h>
-#include <sys/types.h>
 #include <unistd.h>
 
 /**
@@ -165,6 +164,10 @@ static void bftw_heap_bubble_down(struct bftw_cache *cache, struct bftw_dir *dir
 			}
 		}
 
+		if (bftw_heap_check(dir, child)) {
+			break;
+		}
+
 		bftw_heap_move(cache, child, i);
 		i = ci;
 	}
@@ -172,6 +175,22 @@ static void bftw_heap_bubble_down(struct bftw_cache *cache, struct bftw_dir *dir
 	bftw_heap_move(cache, dir, i);
 }
 
+/** Bubble an entry up or down the heap. */
+static void bftw_heap_bubble(struct bftw_cache *cache, struct bftw_dir *dir) {
+	size_t i = dir->heap_index;
+
+	if (i > 0) {
+		size_t pi = (i - 1)/2;
+		struct bftw_dir *parent = cache->heap[pi];
+		if (!bftw_heap_check(parent, dir)) {
+			bftw_heap_bubble_up(cache, dir);
+			return;
+		}
+	}
+
+	bftw_heap_bubble_down(cache, dir);
+}
+
 /** Increment a bftw_dir's reference count. */
 static void bftw_dir_incref(struct bftw_cache *cache, struct bftw_dir *dir) {
 	++dir->refcount;
@@ -210,7 +229,7 @@ static void bftw_cache_remove(struct bftw_cache *cache, struct bftw_dir *dir) {
 	if (i != size) {
 		struct bftw_dir *end = cache->heap[size];
 		end->heap_index = i;
-		bftw_heap_bubble_down(cache, end);
+		bftw_heap_bubble(cache, end);
 	}
 }
 
@@ -545,20 +564,18 @@ static struct bftw_dir *bftw_queue_pop(struct bftw_queue *queue) {
 
 /** Destroy a bftw_queue. */
 static void bftw_queue_destroy(struct bftw_queue *queue) {
+	assert(queue->size == 0);
 	free(queue->buffer);
 }
 
 /** Call stat() and use the results. */
-static int ftwbuf_stat(struct BFTW *ftwbuf, struct stat *sb) {
-	int ret = fstatat(ftwbuf->at_fd, ftwbuf->at_path, sb, ftwbuf->at_flags);
-	if (ret != 0) {
-		return ret;
+static int ftwbuf_stat(struct BFTW *ftwbuf, struct bfs_stat *sb) {
+	int ret = bfs_stat(ftwbuf->at_fd, ftwbuf->at_path, ftwbuf->at_flags, BFS_STAT_BROKEN_OK, sb);
+	if (ret == 0) {
+		ftwbuf->statbuf = sb;
+		ftwbuf->typeflag = mode_to_typeflag(sb->mode);
 	}
-
-	ftwbuf->statbuf = sb;
-	ftwbuf->typeflag = mode_to_typeflag(sb->st_mode);
-
-	return 0;
+	return ret;
 }
 
 /**
@@ -595,7 +612,7 @@ struct bftw_state {
 	/** The current directory. */
 	struct bftw_dir *current;
 	/** The previous directory. */
-	struct bftw_dir *last;
+	struct bftw_dir *previous;
 	/** The currently open directory. */
 	DIR *dir;
 	/** The current traversal status. */
@@ -608,8 +625,8 @@ struct bftw_state {
 
 	/** Extra data about the current file. */
 	struct BFTW ftwbuf;
-	/** stat() buffer for the current file. */
-	struct stat statbuf;
+	/** bfs_stat() buffer for the current file. */
+	struct bfs_stat statbuf;
 };
 
 /**
@@ -638,7 +655,7 @@ static int bftw_state_init(struct bftw_state *state, const char *root, bftw_fn *
 	}
 
 	state->current = NULL;
-	state->last = NULL;
+	state->previous = NULL;
 	state->dir = NULL;
 	state->status = BFTW_CURRENT;
 
@@ -670,25 +687,25 @@ static int bftw_build_path(struct bftw_state *state) {
 	}
 
 	// Only rebuild the part of the path that changes
-	const struct bftw_dir *last = state->last;
-	while (last && last->depth > dir->depth) {
-		last = last->parent;
+	const struct bftw_dir *base = state->previous;
+	while (base && base->depth > dir->depth) {
+		base = base->parent;
 	}
 
 	// Build the path backwards
 	char *path = state->path;
-	while (dir != last) {
+	while (dir != base) {
 		char *segment = path + dir->nameoff;
 		namelen = dir->namelen;
 		memcpy(segment, dir->name, namelen);
 
-		if (last && last->depth == dir->depth) {
-			last = last->parent;
+		if (base && base->depth == dir->depth) {
+			base = base->parent;
 		}
 		dir = dir->parent;
 	}
 
-	state->last = state->current;
+	state->previous = state->current;
 
 	return 0;
 }
@@ -728,7 +745,7 @@ static void bftw_path_trim(struct bftw_state *state) {
 		if (current->namelen > 1) {
 			// Trim the trailing slash
 			--length;
-			state->last = current->parent;
+			state->previous = current->parent;
 		}
 	}
 	dstresize(&state->path, length);
@@ -836,21 +853,14 @@ static void bftw_init_buffers(struct bftw_state *state, const struct dirent *de)
 	    || ftwbuf->typeflag == BFTW_UNKNOWN
 	    || (ftwbuf->typeflag == BFTW_LNK && follow)
 	    || (ftwbuf->typeflag == BFTW_DIR && (detect_cycles || xdev))) {
-		int ret = ftwbuf_stat(ftwbuf, &state->statbuf);
-		if (ret != 0 && follow && (errno == ENOENT || errno == ENOTDIR)) {
-			// Could be a broken symlink, retry without following
-			ftwbuf->at_flags = AT_SYMLINK_NOFOLLOW;
-			ret = ftwbuf_stat(ftwbuf, &state->statbuf);
-		}
-
-		if (ret != 0) {
+		if (ftwbuf_stat(ftwbuf, &state->statbuf) != 0) {
 			bftw_set_error(state, errno);
 			return;
 		}
 
 		if (ftwbuf->typeflag == BFTW_DIR && detect_cycles) {
-			dev_t dev = ftwbuf->statbuf->st_dev;
-			ino_t ino = ftwbuf->statbuf->st_ino;
+			dev_t dev = ftwbuf->statbuf->dev;
+			ino_t ino = ftwbuf->statbuf->ino;
 			for (const struct bftw_dir *dir = current; dir; dir = dir->parent) {
 				if (dev == dir->dev && ino == dir->ino) {
 					bftw_set_error(state, ELOOP);
@@ -899,10 +909,10 @@ static struct bftw_dir *bftw_add(struct bftw_state *state, const char *name) {
 		return NULL;
 	}
 
-	const struct stat *statbuf = state->ftwbuf.statbuf;
+	const struct bfs_stat *statbuf = state->ftwbuf.statbuf;
 	if (statbuf) {
-		dir->dev = statbuf->st_dev;
-		dir->ino = statbuf->st_ino;
+		dir->dev = statbuf->dev;
+		dir->ino = statbuf->ino;
 	}
 
 	return dir;
@@ -958,7 +968,7 @@ static int bftw_gc(struct bftw_state *state, struct bftw_dir *dir, bool invoke_c
 		dir = parent;
 	}
 
-	state->last = dir;
+	state->previous = dir;
 	return ret;
 }
 
@@ -1064,11 +1074,12 @@ int bftw(const char *path, bftw_fn *fn, int nopenfd, enum bftw_flags flags, void
 				break;
 			}
 
-			if (strcmp(de->d_name, ".") == 0 || strcmp(de->d_name, "..") == 0) {
+			const char *name = de->d_name;
+			if (name[0] == '.' && (name[1] == '\0' || (name[1] == '.' && name[2] == '\0'))) {
 				continue;
 			}
 
-			if (bftw_path_concat(&state, de->d_name) != 0) {
+			if (bftw_path_concat(&state, name) != 0) {
 				goto fail;
 			}
 
@@ -1092,14 +1103,14 @@ int bftw(const char *path, bftw_fn *fn, int nopenfd, enum bftw_flags flags, void
 			}
 
 			if (state.ftwbuf.typeflag == BFTW_DIR) {
-				const struct stat *statbuf = state.ftwbuf.statbuf;
+				const struct bfs_stat *statbuf = state.ftwbuf.statbuf;
 				if ((flags & BFTW_XDEV)
 				    && statbuf
-				    && statbuf->st_dev != state.current->dev) {
+				    && statbuf->dev != state.current->dev) {
 					continue;
 				}
 
-				if (bftw_push(&state, de->d_name) != 0) {
+				if (bftw_push(&state, name) != 0) {
 					goto fail;
 				}
 			}
