@@ -72,14 +72,15 @@ make_perms "$TMP/perms"
 
 # Creates a file+directory structure with various symbolic and hard links
 function make_links() {
-    touchp "$1/a"
-    ln -s a "$1/b"
-    ln "$1/a" "$1/c"
-    mkdir -p "$1/d/e/f"
-    ln -s ../../d "$1/d/e/g"
-    ln -s d/e "$1/h"
-    ln -s q "$1/d/e/i"
-    ln -s b/c "$1/j"
+    touchp "$1/file"
+    ln -s file "$1/symlink"
+    ln "$1/file" "$1/hardlink"
+    ln -s nowhere "$1/broken"
+    ln -s symlink/file "$1/notdir"
+    ln -s loop "$1/loop"
+    mkdir -p "$1/deeply/nested/dir"
+    ln -s ../../deeply "$1/deeply/nested/loop"
+    ln -s deeply/nested/loop/nested "$1/skip"
 }
 make_links "$TMP/links"
 
@@ -120,7 +121,7 @@ function make_deep() {
     name="${name}${name}${name}${name}"
     name="${name:0:255}"
 
-    for i in {1..8}; do
+    for i in {0..9} A B C D E F; do
         (
             mkdir "$1/$i"
             cd "$1/$i"
@@ -128,7 +129,7 @@ function make_deep() {
             # 17 * 256 > 16 * 256 == 4096 == PATH_MAX
             for j in {1..17}; do
                 mkdir "$name"
-                cd "$name"
+                cd "$name" 2>/dev/null
             done
         )
     done
@@ -169,6 +170,7 @@ posix_tests=(
     test_name_trailing_slash
     test_path
     test_newer
+    test_newer_link
     test_links
     test_links_plus
     test_links_minus
@@ -187,6 +189,7 @@ posix_tests=(
     test_size_bytes
     test_exec
     test_exec_plus
+    test_exec_plus_status
     test_exec_plus_semicolon
     test_flag_comma
     test_perm_222
@@ -205,7 +208,11 @@ posix_tests=(
     test_a
     test_o
     test_deep
-    test_deep_strict
+    test_double_negation
+    test_de_morgan_not
+    test_de_morgan_and
+    test_de_morgan_or
+    test_data_flow_type
 )
 
 bsd_tests=(
@@ -214,6 +221,10 @@ bsd_tests=(
     test_X
     test_follow
     test_samefile
+    test_samefile_symlink
+    test_H_samefile_symlink
+    test_samefile_broken
+    test_H_samefile_broken
     test_name_slash
     test_name_slashes
     test_iname
@@ -247,10 +258,13 @@ bsd_tests=(
     test_depth_depth_n
     test_depth_depth_n_plus
     test_depth_depth_n_minus
+    test_depth_overflow
     test_gid_name
     test_uid_name
     test_mnewer
     test_H_mnewer
+    test_perm_222_plus
+    test_perm_644_plus
     test_size_T
     test_quit
     test_quit_child
@@ -264,6 +278,7 @@ bsd_tests=(
     test_nouser
     test_exit
     test_printx
+    test_data_flow_depth
 )
 
 gnu_tests=(
@@ -284,6 +299,10 @@ gnu_tests=(
     test_P_slash
     test_follow
     test_samefile
+    test_samefile_symlink
+    test_H_samefile_symlink
+    test_samefile_broken
+    test_H_samefile_broken
     test_xtype_l
     test_xtype_f
     test_L_xtype_l
@@ -313,6 +332,7 @@ gnu_tests=(
     test_flag_weird_names
     test_follow_comma
     test_fprint
+    test_fprint_duplicate
     test_double_dash
     test_flag_double_dash
     test_ignore_readdir_race
@@ -357,25 +377,32 @@ gnu_tests=(
     test_precedence
     test_and_purity
     test_or_purity
+    test_not_reachability
+    test_comma_reachability
+    test_print_error
+    test_fprint_error
 )
 
 bfs_tests=(
     test_type_multi
     test_xtype_multi
+    test_xtype_reorder
     test_perm_symbolic_trailing_comma
     test_perm_symbolic_double_comma
     test_perm_symbolic_missing_action
     test_perm_leading_plus_symbolic
-    test_perm_octal_plus
     test_execdir_plus
     test_hidden
     test_nohidden
+    test_printf_w
     test_path_flag_expr
     test_path_expr_flag
     test_flag_expr_path
     test_expr_flag_path
     test_expr_path_flag
     test_colors
+    test_deep_strict
+    test_deep_stricter
 )
 
 BSD=yes
@@ -450,11 +477,30 @@ fi
 function bfs_verbose() {
     if [ "$VERBOSE" ]; then
         if [ -t 3 ]; then
-            printf '\033[1;32m%s\033[0m ' $BFS >&3
-            echo "$@" >&3
+            printf '\033[1;32m%q\033[0m ' $BFS >&3
+
+            local expr_started=
+            for arg; do
+                if [[ $arg == -[A-Z]* ]]; then
+                    printf '\033[1;36m%q\033[0m ' "$arg" >&3
+                elif [[ $arg == [\(!] || $arg == -[ao] || $arg == -and || $arg == -or || $arg == -not ]]; then
+                    expr_started=yes
+                    printf '\033[1;31m%q\033[0m ' "$arg" >&3
+                elif [[ $expr_started && $arg == [\),] ]]; then
+                    printf '\033[1;31m%q\033[0m ' "$arg" >&3
+                elif [[ $arg == -?* ]]; then
+                    expr_started=yes
+                    printf '\033[1;34m%q\033[0m ' "$arg" >&3
+                elif [ "$expr_started" ]; then
+                    printf '\033[1m%q\033[0m ' "$arg" >&3
+                else
+                    printf '\033[1;35m%q\033[0m ' "$arg" >&3
+                fi
+            done
         else
-            echo $BFS "$@" >&3
+            printf '%q ' $BFS "$@" >&3
         fi
+        printf '\n' >&3
     fi
 }
 
@@ -479,12 +525,18 @@ function bfs_diff() (
 )
 
 function closefrom() {
-    for fd in /dev/fd/*; do
+    if [ -d /proc/self/fd ]; then
+        local fds=/proc/self/fd
+    else
+        local fds=/dev/fd
+    fi
+
+    for fd in "$fds"/*; do
         if [ ! -e "$fd" ]; then
             continue
         fi
 
-        fd="${fd##*/}"
+        local fd="${fd##*/}"
         if [ "$fd" -ge "$1" ]; then
             eval "exec ${fd}<&-"
         fi
@@ -623,6 +675,10 @@ function test_newer() {
     bfs_diff times -newer times/a
 }
 
+function test_newer_link() {
+    bfs_diff times -newer times/l
+}
+
 function test_anewer() {
     bfs_diff times -anewer times/a
 }
@@ -640,23 +696,23 @@ function test_links_minus() {
 }
 
 function test_P() {
-    bfs_diff -P links/d/e/f
+    bfs_diff -P links/deeply/nested/dir
 }
 
 function test_P_slash() {
-    bfs_diff -P links/d/e/f/
+    bfs_diff -P links/deeply/nested/dir/
 }
 
 function test_H() {
-    bfs_diff -H links/d/e/f
+    bfs_diff -H links/deeply/nested/dir
 }
 
 function test_H_slash() {
-    bfs_diff -H links/d/e/f/
+    bfs_diff -H links/deeply/nested/dir/
 }
 
 function test_H_broken() {
-    bfs_diff -H links/d/e/i
+    bfs_diff -H links/broken
 }
 
 function test_H_newer() {
@@ -680,15 +736,31 @@ function test_L_depth() {
 }
 
 function test_samefile() {
-    bfs_diff links -samefile links/a
+    bfs_diff links -samefile links/file
+}
+
+function test_samefile_symlink() {
+    bfs_diff links -samefile links/symlink
+}
+
+function test_H_samefile_symlink() {
+    bfs_diff -H links -samefile links/symlink
+}
+
+function test_samefile_broken() {
+    bfs_diff links -samefile links/broken
+}
+
+function test_H_samefile_broken() {
+    bfs_diff -H links -samefile links/broken
 }
 
 function test_xtype_l() {
-    bfs_diff links -xtype l
+    bfs_diff links -xtype l 2>/dev/null
 }
 
 function test_xtype_f() {
-    bfs_diff links -xtype f
+    bfs_diff links -xtype f 2>/dev/null
 }
 
 function test_L_xtype_l() {
@@ -700,7 +772,14 @@ function test_L_xtype_f() {
 }
 
 function test_xtype_multi() {
-    bfs_diff links -xtype f,d,c
+    bfs_diff links -xtype f,d,c 2>/dev/null
+}
+
+function test_xtype_reorder() {
+    # Make sure -xtype is not reordered in front of anything -- if -xtype runs
+    # before -links 100, it will report an ELOOP error
+    bfs_diff links -links 100 -xtype l
+    invoke_bfs links -links 100 -xtype l
 }
 
 function test_iname() {
@@ -779,12 +858,19 @@ function test_exec_plus() {
     bfs_diff basic -exec "$TESTS/sort-args.sh" '{}' +
 }
 
+function test_exec_plus_status() {
+    # -exec ... {} + should always return true, but if the command fails, bfs
+    # should exit with a non-zero status
+    bfs_diff basic -exec false '{}' + -print
+    ! invoke_bfs basic -exec false '{}' +
+}
+
 function test_exec_plus_semicolon() {
     # POSIX says:
     #     Only a <plus-sign> that immediately follows an argument containing only the two characters "{}"
     #     shall punctuate the end of the primary expression. Other uses of the <plus-sign> shall not be
     #     treated as special.
-    bfs_diff basic -exec "$TESTS/sort-args.sh" foo '{}' bar + baz \;
+    bfs_diff basic -exec echo foo '{}' bar + baz \;
 }
 
 function test_exec_substring() {
@@ -804,7 +890,7 @@ function test_execdir_substring() {
 }
 
 function test_execdir_plus_semicolon() {
-    bfs_diff basic -execdir "$TESTS/sort-args.sh" foo '{}' bar + baz \;
+    bfs_diff basic -execdir echo foo '{}' bar + baz \;
 }
 
 function test_execdir_pwd() {
@@ -849,16 +935,31 @@ function test_follow_comma() {
 }
 
 function test_fprint() {
+    invoke_bfs basic -fprint scratch/test_fprint.out
+    sort -o scratch/test_fprint.out scratch/test_fprint.out
+
     if [ "$UPDATE" ]; then
-        invoke_bfs basic -fprint "$TESTS/test_fprint.out"
-        sort -o "$TESTS/test_fprint.out" "$TESTS/test_fprint.out"
+        cp scratch/test_fprint.out "$TESTS/test_fprint.out"
     else
-        invoke_bfs basic -fprint scratch/test_fprint.out
-        sort -o scratch/test_fprint.out scratch/test_fprint.out
         diff -u scratch/test_fprint.out "$TESTS/test_fprint.out"
     fi
 }
 
+function test_fprint_duplicate() {
+    touchp scratch/test_fprint_duplicate.out
+    ln scratch/test_fprint_duplicate.out scratch/test_fprint_duplicate.hard
+    ln -s test_fprint_duplicate.out scratch/test_fprint_duplicate.soft
+
+    invoke_bfs basic -fprint scratch/test_fprint_duplicate.out -fprint scratch/test_fprint_duplicate.hard -fprint scratch/test_fprint_duplicate.soft
+    sort -o scratch/test_fprint_duplicate.out scratch/test_fprint_duplicate.out
+
+    if [ "$UPDATE" ]; then
+        cp scratch/test_fprint_duplicate.out "$TESTS/test_fprint_duplicate.out"
+    else
+        diff -u scratch/test_fprint_duplicate.out "$TESTS/test_fprint_duplicate.out"
+    fi
+}
+
 function test_double_dash() {
     cd basic
     bfs_diff -- . -type f
@@ -902,6 +1003,10 @@ function test_perm_222_slash() {
     bfs_diff perms -perm /222
 }
 
+function test_perm_222_plus() {
+    bfs_diff perms -perm +222
+}
+
 function test_perm_644() {
     bfs_diff perms -perm 644
 }
@@ -914,6 +1019,10 @@ function test_perm_644_slash() {
     bfs_diff perms -perm /644
 }
 
+function test_perm_644_plus() {
+    bfs_diff perms -perm +644
+}
+
 function test_perm_symbolic() {
     bfs_diff perms -perm a+r,u=wX,g+wX-w
 }
@@ -950,10 +1059,6 @@ function test_perm_leading_plus_symbolic_slash() {
     bfs_diff perms -perm /+rwx
 }
 
-function test_perm_octal_plus() {
-    ! invoke_bfs perms -perm +777 2>/dev/null
-}
-
 function test_permcopy() {
     bfs_diff perms -perm u+rw,g+u-w,o=g
 }
@@ -1069,6 +1174,10 @@ function test_depth_depth_n_minus() {
     bfs_diff basic -depth -depth -2
 }
 
+function test_depth_overflow() {
+    bfs_diff basic -depth -4294967296
+}
+
 function test_gid_name() {
     bfs_diff basic -gid "$(id -gn)"
 }
@@ -1182,29 +1291,34 @@ function test_printf_nul() {
     fi
 }
 
+function test_printf_w() {
+    # Birth times may not be supported, so just check that %w/%W can be parsed
+    bfs_diff times -false -printf '%w %WY\n'
+}
+
 function test_fstype() {
     fstype="$(invoke_bfs -printf '%F\n' | head -n1)"
     bfs_diff basic -fstype "$fstype"
 }
 
 function test_path_flag_expr() {
-    bfs_diff links/h -H -type l
+    bfs_diff links/skip -H -type l
 }
 
 function test_path_expr_flag() {
-    bfs_diff links/h -type l -H
+    bfs_diff links/skip -type l -H
 }
 
 function test_flag_expr_path() {
-    bfs_diff -H -type l links/h
+    bfs_diff -H -type l links/skip
 }
 
 function test_expr_flag_path() {
-    bfs_diff -type l -H links/h
+    bfs_diff -type l -H links/skip
 }
 
 function test_expr_path_flag() {
-    bfs_diff -type l links/h -H
+    bfs_diff -type l links/skip -H
 }
 
 function test_parens() {
@@ -1254,17 +1368,24 @@ function test_colors() {
 function test_deep() {
     closefrom 4
 
-    ulimit -n 8
-    bfs_diff deep -mindepth 18
+    ulimit -n 16
+    bfs_diff deep -mindepth 18 -exec /bin/bash -c 'echo "${1:0:6}/.../${1##*/} (${#1})"' /bin/bash '{}' \;
 }
 
 function test_deep_strict() {
-    # Low ulimit would interfere with the dup()'d stdout for verbose logging
     closefrom 4
 
     # Not even enough fds to keep the root open
+    ulimit -n 7
+    bfs_diff deep -mindepth 18 -exec /bin/bash -c 'echo "${1:0:6}/.../${1##*/} (${#1})"' /bin/bash '{}' \;
+}
+
+function test_deep_stricter() {
+    closefrom 4
+
+    # So few fds that pipe() fails in the -exec implementation
     ulimit -n 6
-    bfs_diff deep -mindepth 18
+    bfs_diff deep -mindepth 18 -exec /bin/bash -c 'echo "${1:0:6}/.../${1##*/} (${#1})"' /bin/bash '{}' \;
 }
 
 function test_exit() {
@@ -1295,6 +1416,50 @@ function test_or_purity() {
     bfs_diff basic -name '*' -o -print
 }
 
+function test_double_negation() {
+    bfs_diff basic \! \! -name 'foo'
+}
+
+function test_not_reachability() {
+    bfs_diff basic -print \! -quit -print
+}
+
+function test_comma_reachability() {
+    bfs_diff basic -print -quit , -print
+}
+
+function test_de_morgan_not() {
+    bfs_diff basic \! \( -name 'foo' -o \! -type f \)
+}
+
+function test_de_morgan_and() {
+    bfs_diff basic \( \! -name 'foo' -a \! -type f \)
+}
+
+function test_de_morgan_or() {
+    bfs_diff basic \( \! -name 'foo' -o \! -type f \)
+}
+
+function test_data_flow_depth() {
+    bfs_diff basic -depth +1 -depth -4
+}
+
+function test_data_flow_type() {
+    bfs_diff basic \! \( -type f -o \! -type f \)
+}
+
+function test_print_error() {
+    if [ -e /dev/full ]; then
+        ! invoke_bfs basic -maxdepth 0 >/dev/full 2>/dev/null
+    fi
+}
+
+function test_fprint_error() {
+    if [ -e /dev/full ]; then
+        ! invoke_bfs basic -maxdepth 0 -fprint /dev/full 2>/dev/null
+    fi
+}
+
 if [ -t 1 -a ! "$VERBOSE" ]; then
     in_place=yes
 fi
@@ -1318,7 +1483,11 @@ for test in ${!run_*}; do
         ((++passed))
     else
         ((++failed))
-        echo "$test failed!"
+        if [ "$in_place" ]; then
+            printf '\r\033[J%s\n' "$test failed!"
+        else
+            echo "$test failed!"
+        fi
     fi
 done
 
