@@ -119,6 +119,24 @@ int dup_cloexec(int fd) {
 #endif
 }
 
+int pipe_cloexec(int pipefd[2]) {
+#if __linux__ || BSD
+	return pipe2(pipefd, O_CLOEXEC);
+#else
+	if (pipe(pipefd) != 0) {
+		return -1;
+	}
+
+	if (fcntl(pipefd[0], F_SETFD, FD_CLOEXEC) == -1 || fcntl(pipefd[1], F_SETFD, FD_CLOEXEC) == -1) {
+		close(pipefd[1]);
+		close(pipefd[0]);
+		return -1;
+	}
+
+	return 0;
+#endif
+}
+
 char *xregerror(int err, const regex_t *regex) {
 	size_t len = regerror(err, regex, NULL, 0);
 	char *str = malloc(len);
@@ -226,6 +244,24 @@ const char *xbasename(const char *path) {
 	return i;
 }
 
+int xfaccessat(int fd, const char *path, int amode) {
+	int ret = faccessat(fd, path, amode, 0);
+
+#ifdef AT_EACCESS
+	// Some platforms, like Hurd, only support AT_EACCESS.  Other platforms,
+	// like Android, don't support AT_EACCESS at all.
+	if (ret != 0 && (errno == EINVAL || errno == ENOTSUP)) {
+		ret = faccessat(fd, path, amode, AT_EACCESS);
+	}
+#endif
+
+	return ret;
+}
+
+bool is_nonexistence_error(int error) {
+	return error == ENOENT || errno == ENOTDIR;
+}
+
 enum bftw_typeflag mode_to_typeflag(mode_t mode) {
 	switch (mode & S_IFMT) {
 #ifdef S_IFBLK
